from torch.utils.data import Dataset as TorchDataset
import torch
from functools import partial
from datasets import load_dataset, ClassLabel, IterableDataset
from datasets import Dataset as HFDataset
from transformers import PreTrainedTokenizer
import numpy as np
from typing import Optional, Union, List, Tuple

DATASET_STREAMING_BUFFER_SIZE = 100000
TEXT_SEPARATOR = ' [SEP] '


class EmptyDatasetError(Exception):
    pass


def gen_from_iterable_dataset(iterable_ds):
    yield from iterable_ds


def concat_columns(inputs, column_list: tuple):
    if isinstance(inputs, str):
        return TEXT_SEPARATOR.join([inputs[col] for col in column_list])

    return [TEXT_SEPARATOR.join([row[col] for col in column_list]) for row in inputs]


class Dataset(TorchDataset):

    def __init__(
            self,
            dataset: Union[HFDataset, IterableDataset],
            text_col: Union[str, Tuple[str]],
            label_col: str,
            is_regression: bool,
            metadata: Optional[dict] = None,
    ):
        self.dataset = dataset
        self.text_col = text_col
        self.label_col = label_col
        self.is_regression = is_regression

        self.dataset_len = len(self.dataset)

        if self.dataset_len == 0:
            raise EmptyDatasetError(f'Dataset is empty.')

        label_features = self.dataset.features[label_col]

        self.has_string_labels = label_features.dtype == 'string'
        if self.has_string_labels:
            if isinstance(dataset, IterableDataset):
                label_list = sorted(list(set([example[label_col] for example in self.dataset])))
            else:
                label_list = sorted(list(set(self.dataset[label_col])))
            self.label_dim = len(label_list)
            self.class_label = ClassLabel(num_classes=self.label_dim, names=label_list)
        elif 'float' in label_features.dtype:
            self.label_dim = 1
            self.class_label = None
        else:
            try:
                self.label_dim = label_features.num_classes
            except:
                self.label_dim = np.max(self.dataset[label_col]) + 1
            self.class_label = None

        self.metadata = metadata

    @classmethod
    def from_hugging_face(
            cls,
            name: str,
            split: str,
            text_col: Union[str, List[str]],
            label_col: str,
            is_regression: bool,
            subset: Optional[str] = None,
            num_examples: Optional[int] = None,
            seed: Optional[int] = None,
            streaming: bool = False
            # load_locally: bool = True
    ) -> "Dataset":
        # if load_locally:
        #     dataset_path = get_dataset_filepath(dataset_name=dataset_name,
        #                                         split=split,
        #                                         max_num_examples=max_num_examples,
        #                                         seed=seed)
        #     if os.path.isdir(dataset_path):
        #         return load_from_disk(dataset_path)

        if subset is None:
            dataset = load_dataset(name, split=split, streaming=streaming)  # [12250:12750]
        else:
            dataset = load_dataset(name, subset, split=split, streaming=streaming)  # [12250:12750]

        cols_to_keep = text_col + [label_col] if isinstance(text_col, list) else [text_col, label_col]
        dataset = dataset.select_columns(cols_to_keep)

        task_type = "regression" if is_regression else "classification"
        if task_type == 'classification':
            dataset = dataset.filter(lambda example: example[label_col] != -1)

        if num_examples is not None:
            if streaming:
                dataset = dataset.shuffle(seed=seed, buffer_size=DATASET_STREAMING_BUFFER_SIZE).take(num_examples)
                dataset = HFDataset.from_generator(
                    partial(gen_from_iterable_dataset, dataset),
                    features=dataset.features
                )
            else:
                num_examples = min(len(dataset), num_examples)
                dataset = dataset.shuffle(seed=seed)
                dataset = dataset.select(range(num_examples))
        else:
            if streaming:
                dataset = HFDataset.from_generator(
                    partial(gen_from_iterable_dataset, dataset),
                    features=dataset.features
                )

        metadata = {
            "streamed": streaming,
            "seed": seed
        }

        return Dataset(
            dataset=dataset,
            text_col=text_col,
            label_col=label_col,
            is_regression=is_regression,
            metadata=metadata,
        )

    def __getitem__(self, idx):
        return self.dataset[idx]

    def __len__(self):
        return self.dataset_len
    #
    # def save_locally(self):
    #     filepath = get_dataset_filepath(dataset_name=self.dataset_name,
    #                                     split=self.split,
    #                                     max_num_examples=self.max_num_examples,
    #                                     seed=self.seed)
    #
    #     self.dataset.save_to_disk(filepath)

    def collate_fn(
            self,
            rows: dict,
            tokenizer: PreTrainedTokenizer,
            max_length: int = 128,
            return_token_type_ids: bool = False
    ):
        texts = self._preprocess_texts(rows)
        labels = self._preprocess_labels(rows)

        tokenized = tokenizer(texts,
                              padding='max_length',
                              truncation=True,
                              max_length=max_length,
                              return_tensors='pt',
                              return_token_type_ids=return_token_type_ids)
        if return_token_type_ids:
            return tokenized.data['input_ids'], \
                   tokenized.data['attention_mask'], \
                   tokenized.data['token_type_ids'], \
                   torch.tensor(labels)

        return tokenized.data['input_ids'], tokenized.data['attention_mask'], torch.tensor(labels)

    def _preprocess_texts(self, rows):
        if isinstance(self.text_col, (list, tuple)):
            inputs = concat_columns(rows, self.text_col)
        else:
            inputs = [row[self.text_col] for row in rows]

        return inputs

    def _preprocess_labels(self, rows):
        labels = [row[self.label_col] for row in rows]
        if self.has_string_labels:
            labels = [self.class_label.str2int(label) for label in labels]

        return labels

    # def train_test_split(self, train_size=0.8, test_size=None, seed=42, shuffle=True):
    #     if test_size is not None:
    #         train_size = 1 - test_size
    #
    #     train_dataset = copy(self)
    #     test_dataset = copy(self)
    #
    #     split_internal_ds = self.dataset.train_test_split(train_size=train_size, seed=seed,
    #                                                                         shuffle=shuffle)
    #     internal_train_ds = split_internal_ds['train']
    #     internal_test_ds = split_internal_ds['test']
    #
    #     train_dataset.dataset = internal_train_ds
    #     train_dataset.dataset_len = len(internal_train_ds)
    #
    #     test_dataset.dataset = internal_test_ds
    #     test_dataset.dataset_len = len(internal_test_ds)
    #
    #     return train_dataset, test_dataset



from torch.utils.data import SequentialSampler, DataLoader
from .LogME import LogME
import numpy as np
import torch
from .model_utils import get_pooled_output
from transformers import PreTrainedModel, PreTrainedTokenizer
from .ESM import ESM, fetch_esms
from .dataset import Dataset
from tqdm import tqdm
from typing import List
from transformers import AutoModel, AutoTokenizer


def compute_logme_esm_batch(dataset: Dataset,
                            base_model: PreTrainedModel,
                            esms: List[ESM],
                            tokenizer: PreTrainedTokenizer,
                            # regression: bool,
                            batch_size: int = 128,
                            device_name: str = "cpu"):
    sampler = SequentialSampler(dataset)
    dataloader = DataLoader(dataset, sampler=sampler, batch_size=batch_size,
                            collate_fn=lambda x: dataset.collate_fn(x, tokenizer=tokenizer))
    device = torch.device(device_name)
    base_model.to(device)

    regression = dataset.is_regression
    if regression:
        label_dtype = float
    else:
        label_dtype = int

    labels = np.zeros(0, label_dtype)
    esm_embeddings = [[] for _ in range(len(esms))]

    for batch in tqdm(dataloader):
        batch = tuple(t.to(device) for t in batch)
        b_input_ids, b_input_mask, b_labels = batch
        b_labels = b_labels.detach().cpu().numpy().flatten()

        with torch.no_grad():
            batch_base_embeddings = get_pooled_output(base_model, b_input_ids, b_input_mask)
            for i, transformation_net in enumerate(esms):
                batch_transformed_embeddings = transformation_net(batch_base_embeddings).cpu().numpy()
                esm_embeddings[i].append(batch_transformed_embeddings)

        labels = np.append(labels, b_labels, axis=0)

    logme_results = []
    for features in tqdm(esm_embeddings):
        embeddings = np.vstack(features)
        logme_results.append(LogME(regression=regression).fit(embeddings, labels, add_intercept=False))

    return logme_results


def compute_task_ranking(
        dataset: Dataset,
        model_name: str,
        # is_regression: bool
) -> dict[str, float]:
    esms = fetch_esms(model_name)
    repo_ids = list(esms.keys())
    esms = list(esms.values())

    bert_model = AutoModel.from_pretrained(model_name)
    tokenizer = AutoTokenizer.from_pretrained(model_name)

    scores = compute_logme_esm_batch(
        dataset=dataset,
        base_model=bert_model,
        tokenizer=tokenizer,
        esms=esms,
    )

    return {repo_ids[idx]: scores[idx] for idx in np.argsort(scores)[::-1]}



import torch
import torch.nn as nn
import numpy as np
from typing import List, Dict, Optional
from huggingface_hub import PyTorchModelHubMixin, hf_hub_download, HfApi, create_repo, ModelCard, ModelCardData
from tqdm import tqdm
# from . import hf_api


class ESM(nn.Module, PyTorchModelHubMixin):

    def __init__(
            self,
            embedding_dim: int = 768,
            base_model_name: Optional[str] = None,
            task_id: Optional[str] = None,
            optional_layer_dims: Optional[List[int]] = None,
            metadata: Dict[str, str] = None):
        super(ESM, self).__init__()
        if isinstance(optional_layer_dims, int):
            optional_layer_dims = [optional_layer_dims]
        if optional_layer_dims is None:
            self.layer_dims = [embedding_dim, embedding_dim]
            # layers = [nn.Linear(EMBEDDING_SIZE, EMBEDDING_SIZE)]
        else:
            self.layer_dims = [embedding_dim] + optional_layer_dims + [embedding_dim]

        self.embedding_dim = embedding_dim
        self.relu = nn.ReLU()

        # layers = [nn.Linear(input_dim, output_dim) for input_dim, output_dim in zip(dims, dims[1:])]
        linear_layers = [nn.Linear(input_dim, output_dim) for input_dim, output_dim in zip(self.layer_dims, self.layer_dims[1:])]

        layers = []
        for ll in linear_layers:
            layers += [ll, self.relu]
        layers = layers[:-1]

        # TODO: Fix bottleneck idx after relu insertion
        self.bottleneck_idx = np.argmin(self.layer_dims[1:]) + 1
        self.bottleneck_dim = np.min(self.layer_dims)
        self.sequential = nn.Sequential(*layers)

        self.base_model_name = "" if base_model_name is None else base_model_name
        self.task_id = "" if task_id is None else task_id
        self.metadata = metadata

    def publish(
            self,
            repo_id: str,
            developers: Optional[str] = None
    ) -> None:
        create_repo(repo_id=repo_id, exist_ok=True)
        self.push_to_hub(repo_id=repo_id)

        card_data = ModelCardData(language='en', license='mit', library_name='keras',
                                  tags=["embedding_space_map", f"BaseLM:{self.base_model_name}"])

        model_card_kwargs = {}
        if developers:
            model_card_kwargs["developers"] = developers

        card = ModelCard.from_template(
            card_data,
            model_id=self.task_id,
            model_description="ESM",
            datasets=[self.task_id],
            **model_card_kwargs
        )
        card.push_to_hub(repo_id)

    @classmethod
    def from_disk(
            cls,
            filepath: str,
            device_name: str = "cpu",
    ) -> "ESM":

        device = torch.device(device_name)
        state_dict = torch.load(filepath, map_location=device)
        embedding_dim = state_dict['sequential.0.weight'].shape[1]

        esm = ESM(embedding_dim=embedding_dim)
        esm.tags = ["embeddings_space_map_test"]
        esm.load_state_dict(state_dict)

        return esm

    @classmethod
    def from_hugging_face(cls, repo_id: str) -> "ESM":
        esm = ESM.from_disk(hf_hub_download(repo_id, filename="model.safetensors"))
        esm.repo_id = repo_id

        return esm


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.sequential(x)

    def forward_bottleneck(self, x: torch.Tensor) -> torch.Tensor:
        return self.sequential[:self.bottleneck_idx](x)

    @property
    def bottleneck_model(self) -> nn.Sequential:
        return self.sequential[:self.bottleneck_idx]


def find_esm_repo_ids(model_name: str) -> List[str]:
    hf_api = HfApi()
    model_infos = hf_api.list_models(filter=["embedding_space_map", f"BaseLM:{model_name}"])

    return [model_info.id for model_info in model_infos]


def fetch_esms(model_name: str) -> dict[str, "ESM"]:
    repo_ids = find_esm_repo_ids(model_name=model_name)

    esms = {}
    for repo_id in tqdm(repo_ids):
        try:
            esms[repo_id] = ESM.from_pretrained(repo_id)
        except Exception as e:
            pass

    return esms



import os
from huggingface_hub import HfApi, ModelCardData, ModelCard
from src.hfselect import test_fetch_esms, ESM
import sys
import huggingface_hub

print(huggingface_hub.__version__)

print(sys.version)


task_ids = [
    "glue_qqp",
    "guardian_authorship_cross_genre_4"
]
filepaths = [
    "",
    "../esms/.pt"
]
hf_api = HfApi()

for task_id in task_ids:
    filepath = f"../esms/{task_id}.pt"
    esm = ESM.from_disk(filepath)

    # esm.extra_tags = [task_id]

    repo_id = f"davidschulte/{task_id}"
    esm.push_to_hub(repo_id=repo_id)

    card_data = ModelCardData(language='en', license='mit', library_name='keras',
                              tags=["embedding_space_map", "base_bert-base-uncased"])
    card = ModelCard.from_template(
        card_data,
        model_id=task_id,
        model_description="ESM",
        developers="David Schulte",
    )
    card.push_to_hub(repo_id)

    esm = ESM.from_pretrained(repo_id)


esms = test_fetch_esms()
print(esms)

# esms = test_fetch_esms()
#
# print(esms)



from hfselect import Dataset, compute_task_ranking

dataset = Dataset.from_hugging_face(name="glue",
                                    split="train",
                                    text_col=["premise", "hypothesis"],
                                    label_col="label",
                                    is_regression=False,
                                    subset="mnli",
                                    num_examples=1000,
                                    seed=42)

task_ranking = compute_task_ranking(
    dataset=dataset,
    model_name="bert-base-multilingual-uncased"
)

print(task_ranking)


import numpy as np
from torch.utils.data import Dataset as TorchDataset
from .dataset import Dataset
from transformers import PreTrainedModel, PreTrainedTokenizer
from torch.utils.data import SequentialSampler, DataLoader
import os
from typing import Optional
from tqdm import tqdm
from .model_utils import get_pooled_output
import torch


class EmbeddingDataset(TorchDataset):

    def __init__(self, x, y):
        self.x = x
        self.y = y

        assert len(x) == len(y)

        self.num_rows = len(self.x)

    @classmethod
    def from_disk(cls, filepath):
        x = np.loadtxt(os.path.join(filepath, 'standard_embeddings.csv'))
        y = np.loadtxt(os.path.join(filepath, 'trained_embeddings.csv'))

        return EmbeddingDataset(x, y)

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]

    def __len__(self):
        return self.num_rows


def create_embedding_dataset(
        dataset: Dataset,
        base_model: PreTrainedModel,
        tuned_model: PreTrainedModel,
        tokenizer: PreTrainedTokenizer,
        device_name: str = "cpu",
        output_path: Optional[str] = None,
        batch_size: int = 128,
        overwrite: bool = True
) -> "EmbeddingDataset":
    os.makedirs(output_path, exist_ok=True)
    standard_embeddings_filepath = os.path.join(output_path, f'standard_embeddings.csv')
    trained_embeddings_filepath = os.path.join(output_path, f'trained_embeddings.csv')
    if os.path.isfile(standard_embeddings_filepath) and os.path.isfile(
            trained_embeddings_filepath) and not overwrite:
        print("Found embeddings.")
        return EmbeddingDataset.from_disk(output_path)

    for embedding_filepath in [standard_embeddings_filepath, trained_embeddings_filepath]:
        if os.path.exists(embedding_filepath):
            os.remove(embedding_filepath)

    device = torch.device(device_name)

    base_model.to(device)
    tuned_model.to(device)

    base_model.eval()
    tuned_model.eval()
    print('Loading models complete!')

    sampler = SequentialSampler(dataset)
    dataloader = DataLoader(dataset,
                            sampler=sampler,
                            batch_size=batch_size,
                            collate_fn=lambda x: dataset.collate_fn(x, tokenizer=tokenizer))

    for step, batch in enumerate(tqdm(dataloader)):
        batch = tuple(t.to(device) for t in batch)
        b_input_ids, b_input_mask, _ = batch

        with torch.no_grad():
            trained_embeddings = get_pooled_output(tuned_model, b_input_ids, b_input_mask).cpu().numpy()
            standard_embeddings = get_pooled_output(base_model, b_input_ids,
                                                    b_input_mask).cpu().numpy()

        with open(standard_embeddings_filepath, "ab") as f:
            np.savetxt(f, standard_embeddings)

        with open(trained_embeddings_filepath, "ab") as f:
            np.savetxt(f, trained_embeddings)

    return EmbeddingDataset(standard_embeddings, trained_embeddings)



import torch
from transformers import PreTrainedModel, PreTrainedTokenizer, AutoTokenizer, \
    BertModel, BertConfig,  BertForSequenceClassification, BertTokenizer,\
    RobertaModel, RobertaConfig, RobertaForSequenceClassification, RobertaTokenizer,\
    DistilBertModel, DistilBertConfig, DistilBertForSequenceClassification, DistilBertTokenizer, logging
from copy import deepcopy

SEQUENCE_CLASSIFICATION_MODEL_ARGS = {
    'output_attentions': False,
    'output_hidden_states': False
}

MODELS = {
    'bert': {
        'base_model': BertModel,
        'sequence_classification_model': BertForSequenceClassification,
        'tokenizer': BertTokenizer,
        'config': BertConfig,
        'pretrained_name': 'bert-base-uncased',
        'base_model_attribute_name': 'bert',
        'embedding_size': 768,
        'pooling_method': 'bert_pooling'
    },
    'bert_multilang': {
        'base_model': BertModel,
        'sequence_classification_model': BertForSequenceClassification,
        'tokenizer': BertTokenizer,
        'config': BertConfig,
        'pretrained_name': 'bert-base-multilingual-uncased',
        'base_model_attribute_name': 'bert',
        'embedding_size': 768,
        'pooling_method': 'bert_pooling'
    },
    'bert-tiny': {
        'base_model': BertModel,
        'sequence_classification_model': BertForSequenceClassification,
        'tokenizer': BertTokenizer,
        'config': BertConfig,
        'pretrained_name': 'prajjwal1/bert-tiny',
        'base_model_attribute_name': 'bert',
        'embedding_size': 128,
        'pooling_method': 'bert_pooling'
    },
    'roberta-base': {
        'base_model': RobertaModel,
        'sequence_classification_model': RobertaForSequenceClassification,
        'tokenizer': RobertaTokenizer,
        'config': RobertaConfig,
        'pretrained_name': 'roberta-base',
        'base_model_attribute_name': 'roberta',
        'embedding_size': 768,
        'pooling_method': 'roberta_pooling'
    },
    'distilbert': {
        'base_model': DistilBertModel,
        'sequence_classification_model': DistilBertForSequenceClassification,
        'tokenizer': DistilBertTokenizer,
        'config': DistilBertConfig,
        'pretrained_name': 'distilbert-base-uncased',
        'base_model_attribute_name': 'distilbert',
        'embedding_size': 768,
        'pooling_method': 'distilbert_pooling'
    },
    'sentence_mpnet': {
        'base_model': BertModel,
        'sequence_classification_model': BertForSequenceClassification,
        'tokenizer': AutoTokenizer,
        'config': BertConfig,
        'pretrained_name': 'sentence-transformers/all-mpnet-base-v2',
        'base_model_attribute_name': 'bert',
        'embedding_size': 768,
        'pooling_method': 'sbert_pooling'
    },
    'sentence_minilm': {
        'base_model': BertModel,
        'sequence_classification_model': BertForSequenceClassification,
        'tokenizer': AutoTokenizer,
        'config': BertConfig,
        'pretrained_name': 'sentence-transformers/all-MiniLM-L6-v2',
        'base_model_attribute_name': 'bert',
        'embedding_size': 384,
        'pooling_method': 'sbert_pooling'
    },
    'sentence_minilm_multilang': {
        'base_model': BertModel,
        'sequence_classification_model': BertForSequenceClassification,
        'tokenizer': AutoTokenizer,
        'config': BertConfig,
        'pretrained_name': 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2',
        'base_model_attribute_name': 'bert',
        'embedding_size': 384,
        'pooling_method': 'sbert_pooling'
    }
}


def create_base_model(model_name):
    logging.set_verbosity_error()
    base_model_class: PreTrainedModel = MODELS[model_name]['base_model']
    pretrained_name = MODELS[model_name]['pretrained_name']

    return base_model_class.from_pretrained(pretrained_name)


# def create_sequence_classification_model(num_labels, model_name=MODEL_NAME):
#     logging.set_verbosity_error()
#     sequence_classification_model_class: PreTrainedModel = MODELS[model_name]['sequence_classification_model']
#     pretrained_name = MODELS[model_name]['pretrained_name']
#
#     return sequence_classification_model_class.from_pretrained(pretrained_name,
#                                                                num_labels=num_labels,
#                                                                **SEQUENCE_CLASSIFICATION_MODEL_ARGS)


def create_tokenizer(model_name):
    tokenizer_class: PreTrainedTokenizer = MODELS[model_name]['tokenizer']
    pretrained_name = MODELS[model_name]['pretrained_name']

    return tokenizer_class.from_pretrained(pretrained_name)
#
#
# def load_sequence_classification_model_from_dir(model_dir, model_name=MODEL_NAME):
#     sequence_classification_model_class: PreTrainedModel = MODELS[model_name]['sequence_classification_model']
#
#     return sequence_classification_model_class.from_pretrained(model_dir)
#
#
# def reset_head_layer(model: PreTrainedModel, num_labels, model_name=MODEL_NAME):
#     logging.set_verbosity_error()
#     # model_name = model.base_model_prefix
#     config_class = MODELS[model_name]['config']
#     pretrained_name = MODELS[model_name]['pretrained_name']
#     sequence_classification_model_class = MODELS[model_name]['sequence_classification_model']
#     base_model_attribute_name = MODELS[model_name]['base_model_attribute_name']
#     new_model_config = config_class.from_pretrained(pretrained_name, num_labels=num_labels)
#     new_model = sequence_classification_model_class(new_model_config)
#     # new_model = sequence_classification_model_class(model.config)
#     setattr(new_model, base_model_attribute_name, deepcopy(getattr(model, base_model_attribute_name)))
#
#     return new_model


def get_pooled_output(
        base_model: PreTrainedModel,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor):
    # pooling_method = MODELS[MODEL_NAME]['pooling_method']

    if isinstance(base_model, BertModel):
        if "sentence-transformers" in base_model.name_or_path:
            token_embeddings = base_model(input_ids, attention_mask=attention_mask)[0]  # First element of model_output contains all token embeddings
            input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

        return base_model(input_ids, attention_mask=attention_mask)[1]

    elif isinstance(base_model, DistilBertModel):
        return base_model(input_ids, attention_mask=attention_mask)[0][:, 0]

    elif isinstance(base_model, RobertaModel):
        return base_model(input_ids, attention_mask=attention_mask)[0][:, 0, :]

    else:
        raise Exception

    # if pooling_method == 'bert_pooling':
    #     return base_model(input_ids, attention_mask=attention_mask)[1]
    # elif pooling_method == 'distilbert_pooling':
    #     return base_model(input_ids, attention_mask=attention_mask)[0][:, 0]
    # elif pooling_method == 'sbert_pooling':
    #     token_embeddings = base_model(input_ids, attention_mask=attention_mask)[0]  # First element of model_output contains all token embeddings
    #     input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
    #     return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    # elif pooling_method == 'roberta_pooling':
    #     return base_model(input_ids, attention_mask=attention_mask)[0][:, 0, :]
    # else:
    #     raise Exception

    # if isinstance(base_model, BertModel):
    #     return base_model(input_ids, attention_mask=attention_mask)[1]
    # elif isinstance(base_model, DistilBertModel):
    #     return base_model(input_ids, attention_mask=attention_mask)[0][:, 0]


# def get_base_model(sequence_classification_model, model_name=MODEL_NAME):
#     base_model_attribute_name = MODELS[model_name]['base_model_attribute_name']
#
#     return getattr(sequence_classification_model, base_model_attribute_name)



from abc import abstractmethod
from ESM import ESM
from transformers import get_linear_schedule_with_warmup, PreTrainedModel
import torch
from torch import nn
from torch.nn import MSELoss
from torch.optim import AdamW
from src.hfselect.model_utils import create_sequence_classification_model, get_base_model
import os
from typing import Optional
import time
import json
from tqdm import tqdm
from datetime import datetime
from torch.utils.data import SequentialSampler, DataLoader
from .embedding_dataset import EmbeddingDataset, create_embedding_dataset
from .dataset import Dataset


class Trainer:
    def __init__(self,
                 model: Optional[nn.Module] = None,
                 optimizer=None,
                 learning_rate=0.001,
                 weight_decay=0.01,
                 scheduler=None,
                 device: str = "cpu"):
        self.output_dim = output_dim
        self.model = self._create_model() if model is None else model
        self.optimizer = self._create_optimizer(self.model, weight_decay, learning_rate=learning_rate)\
            if optimizer is None else optimizer
        # self.scheduler = self._create_scheduler(self.optimizer, num_train_steps=num_train_steps) if scheduler is None \
        #     else scheduler
        self.scheduler = scheduler

        if device != "cpu" and torch.cuda.is_available():
            self.device = torch.device(device) if torch.cuda.is_available() else torch.device("cpu")
        else:
            self.device = "cpu"
        # print(self.device)
        self.model.to(self.device)

        self.total_loss = 0
        self.num_train_examples = 0

    @abstractmethod
    def train_step(self, *args, **kwargs):
        pass

    @abstractmethod
    def _create_model(self):
        pass

    @staticmethod
    def _create_optimizer(model, weight_decay, learning_rate=0.001):
        return AdamW(model.parameters(), lr=learning_rate, weight_decay=weight_decay)

    @staticmethod
    def _create_scheduler(optimizer, num_train_steps):
        return get_linear_schedule_with_warmup(optimizer,
                                               num_warmup_steps=0,
                                               num_training_steps=num_train_steps)

    def reset_loss(self):
        # print(f'Loss: {self.total_loss}')
        self.total_loss = 0
        self.num_train_examples = 0

    @property
    def avg_loss(self):
        return self.total_loss / self.num_train_examples


class TransformerTrainer(Trainer):

    def __init__(self, output_dim, model=None, optimizer=None, weight_decay=0.01, scheduler=None,
                 freeze_base_model=False, device="cpu"):
        super(TransformerTrainer, self).__init__(model=model,
                                                 optimizer=optimizer,
                                                 weight_decay=weight_decay,
                                                 scheduler=scheduler,
                                                 learning_rate=2e-5,
                                                 device=device)

        self.output_dim = output_dim

        self.freeze_base_model = freeze_base_model

        if self.freeze_base_model:
            base_model = get_base_model(self.model)
            for param in base_model.parameters():
                param.requires_grad = False

    def train_step(self, batch):
        batch = tuple(b.to(self.device) for b in batch)
        b_input_ids, b_input_mask, b_labels = batch

        self.model.train()
        self.model.zero_grad()

        outputs = self.model(b_input_ids,
                             # token_type_ids=None,
                             attention_mask=b_input_mask,
                             labels=b_labels)
        loss = outputs[0]
        self.total_loss += loss.item()
        self.num_train_examples += len(b_input_ids)

        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)

        self.optimizer.step()
        self.scheduler.step()

        return loss.item()

    def _create_model(self):
        return create_sequence_classification_model(num_labels=self.output_dim)


class ESMTrainer(Trainer):
    def __init__(self, model=None, model_optional_layer_dims=None,
                 optimizer=None, weight_decay=0.01, learning_rate=0.001, scheduler=None, device="cpu"):

        self.model_optional_layer_dims = model_optional_layer_dims

        super(ESMTrainer, self).__init__(model=model,
                                         optimizer=optimizer,
                                         weight_decay=weight_decay,
                                         scheduler=scheduler,
                                         learning_rate=learning_rate,
                                         device=device)

        self.loss_fct = MSELoss()

    def _create_model(self):
        return ESM(optional_layer_dims=self.model_optional_layer_dims)

    def train_step(self, embeddings_batch):
        self.model.train()

        embeddings_batch = tuple(b.to(self.device) for b in embeddings_batch)
        b_standard_embeddings, b_transferred_embeddings = embeddings_batch

        self.model.zero_grad()
        outputs = self.model(b_standard_embeddings.float())
        # loss = self.loss_fct(outputs, b_labels.float())
        loss = self.loss_fct(outputs, b_transferred_embeddings.float())
        self.total_loss += loss.item()
        self.num_train_examples += len(b_standard_embeddings)
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)
        self.optimizer.step()
        self.scheduler.step()

        return loss.item()

    def train_with_embeddings(self,
                              embeddings_dataset: EmbeddingDataset,
                              output_filepath: str = None,
                              num_epochs: int = 10,
                              batch_size: int = 32,
                              overwrite:bool = False):

        if output_filepath:
            if os.path.isfile(output_filepath) and not overwrite:
                print('Found transformation network on disk.')
                return ESM.from_pretrained(output_filepath)

        sampler = SequentialSampler(embeddings_dataset)
        dataloader = DataLoader(embeddings_dataset, sampler=sampler, batch_size=batch_size)

        num_train_steps = len(dataloader) * num_epochs

        # TODO: Check scheduler if scheduler should always be overwritten
        if self.scheduler is None:
            self.scheduler = self._create_scheduler(optimizer=self.optimizer, num_train_steps=num_train_steps)

        epoch_train_durations = []
        epoch_avg_losses = []
        start_time = time.time()
        for epoch_i in range(num_epochs):

            self.reset_loss()
            with tqdm(dataloader, desc=f'Training: Epoch {epoch_i} / {num_epochs}', unit='batch') as pbar:

                for step, batch in enumerate(pbar):
                    loss = self.train_step(batch)

                    avg_train_loss = loss / batch_size

                    pbar.set_postfix(avg_train_loss=avg_train_loss)

            end_time = time.time()
            epoch_train_durations.append(end_time - start_time)
            start_time = end_time
            epoch_avg_losses.append(self.avg_loss)

        if output_filepath:
            output_dir = os.path.dirname(output_filepath)
            os.makedirs(output_dir, exist_ok=True)
            torch.save(self.model.state_dict(), output_filepath)
            train_info_dict = {
                'training_completed_timestamp': datetime.now().strftime("%m/%d/%Y, %H:%M:%S"),
                'num_epochs': num_epochs,
                'num_train_examples': len(embeddings_dataset),
                'epoch_train_durations': epoch_train_durations,
                'epoch_avg_losses': epoch_avg_losses
            }

            with open(os.path.join(output_dir, 'train_info.json'), 'w') as f:
                json.dump(train_info_dict, f)

            print('Saved model.')

        return self.model

    def train_with_models(self,
                          dataset: Dataset,
                          base_model: PreTrainedModel,
                          tuned_model: PreTrainedModel,
                          model_output_filepath: Optional[str] = None,
                          embeddings_output_filepath: Optional[str] = None,
                          num_epochs: int = 10,
                          train_batch_size: int = 32,
                          embeddings_batch_size: int = 128,
                          device_name: str = "cpu",
                          overwrite_model: bool = False,
                          overwrite_embeddings: bool = False):
        embeddings_dataset = create_embedding_dataset(dataset=dataset,
                                                      base_model=base_model,
                                                      tuned_model=tuned_model,
                                                      batch_size=embeddings_batch_size,
                                                      output_path=embeddings_output_filepath,
                                                      device_name=device_name,
                                                      overwrite=overwrite_embeddings)

        return self.train_with_embeddings(embeddings_dataset=embeddings_dataset,
                                          output_filepath=model_output_filepath,
                                          num_epochs=num_epochs,
                                          batch_size=train_batch_size,
                                          overwrite=overwrite_model)



import warnings

import numpy as np
from numba import njit


@njit
def each_evidence(y_, f, fh, v, s, vh, N, D):
    """
    compute the maximum evidence for each class
    """
    epsilon = 1e-5
    alpha = 1.0
    beta = 1.0
    lam = alpha / beta
    tmp = (vh @ (f @ np.ascontiguousarray(y_)))
    for i in range(100):
        # print(f'{i}/100')
        # should converge after at most 10 steps
        # typically converge after two or three steps
        gamma = (s / (s + lam)).sum()
        # A = v @ np.diag(alpha + beta * s) @ v.transpose() # no need to compute A
        # A_inv = v @ np.diag(1.0 / (alpha + beta * s)) @ v.transpose() # no need to compute A_inv
        m = v @ (tmp * beta / (alpha + beta * s))
        alpha_de = (m * m).sum()
        alpha = gamma / (alpha_de + epsilon)
        beta_de = ((y_ - fh @ m) ** 2).sum()
        beta = (N - gamma) / (beta_de + epsilon)
        new_lam = alpha / beta
        print(np.abs(new_lam - lam))
        if np.abs(new_lam - lam) / lam < 0.01:
            break
        lam = new_lam
    evidence = D / 2.0 * np.log(alpha) \
               + N / 2.0 * np.log(beta) \
               - 0.5 * np.sum(np.log(alpha + beta * s)) \
               - beta / 2.0 * (beta_de + epsilon) \
               - alpha / 2.0 * (alpha_de + epsilon) \
               - N / 2.0 * np.log(2 * np.pi)
    return evidence / N, alpha, beta, m


# use pseudo data to compile the function
# D = 20, N = 50
# f_tmp = np.random.randn(20, 50).astype(np.float64)
# each_evidence(np.random.randint(0, 2, 50).astype(np.float64), f_tmp, f_tmp.transpose(), np.eye(20, dtype=np.float64), np.ones(20, dtype=np.float64), np.eye(20, dtype=np.float64), 50, 20)


@njit
def truncated_svd(x):
    u, s, vh = np.linalg.svd(x.transpose() @ x)
    s = np.sqrt(s)
    u_times_sigma = x @ vh.transpose()
    k = np.sum((s > 1e-10) * 1)  # rank of f
    s = s.reshape(-1, 1)
    s = s[:k]
    vh = vh[:k]
    u = u_times_sigma[:, :k] / s.reshape(1, -1)
    return u, s, vh
# truncated_svd(np.random.randn(20, 10).astype(np.float64))


class LogME(object):
    def __init__(self, regression=False):
        """
            :param regression: whether regression
        """
        # print('logme')
        self.regression = regression
        self.fitted = False
        self.reset()

    def reset(self):
        self.num_dim = 0
        self.alphas = []  # alpha for each class / dimension
        self.betas = []  # beta for each class / dimension
        # self.ms.shape --> [C, D]
        self.ms = []  # m for each class / dimension

    def _fit_icml(self, f: np.ndarray, y: np.ndarray):
        """
        LogME calculation proposed in the ICML 2021 paper
        "LogME: Practical Assessment of Pre-trained Models for Transfer Learning"
        at http://proceedings.mlr.press/v139/you21b.html
        """
        fh = f
        f = f.transpose()
        D, N = f.shape
        v, s, vh = np.linalg.svd(f @ fh, full_matrices=True)

        evidences = []
        self.num_dim = y.shape[1] if self.regression else int(y.max() + 1)
        for i in range(self.num_dim):
            print(f'k: {i}')
            y_ = y[:, i] if self.regression else (y == i).astype(np.float64)
            evidence, alpha, beta, m = each_evidence(y_, f, fh, v, s, vh, N, D)
            evidences.append(evidence)
            self.alphas.append(alpha)
            self.betas.append(beta)
            self.ms.append(m)
        self.ms = np.stack(self.ms)
        return np.mean(evidences)

    def _fit_fixed_point(self, f: np.ndarray, y: np.ndarray):
        """
        LogME calculation proposed in the arxiv 2021 paper
        "Ranking and Tuning Pre-trained Models: A New Paradigm of Exploiting Model Hubs"
        at https://arxiv.org/abs/2110.10545
        """
        N, D = f.shape  # k = min(N, D)
        if N > D: # direct SVD may be expensive
            u, s, vh = truncated_svd(f)
        else:
            u, s, vh = np.linalg.svd(f, full_matrices=False)
        # u.shape = N x k
        # s.shape = k
        # vh.shape = k x D
        s = s.reshape(-1, 1)
        sigma = (s ** 2)

        evidences = []
        self.num_dim = y.shape[1] if self.regression else int(y.max() + 1)
        for i in range(self.num_dim):
            # print(f'k: {i}')
            y_ = y[:, i] if self.regression else (y == i).astype(np.float64)
            y_ = y_.reshape(-1, 1)
            x = u.T @ y_  # x has shape [k, 1], but actually x should have shape [N, 1]
            x2 = x ** 2
            res_x2 = (y_ ** 2).sum() - x2.sum()  # if k < N, we compute sum of xi for 0 singular values directly

            alpha, beta = 1.0, 1.0
            for _ in range(50):
                t = alpha / beta
                gamma = (sigma / (sigma + t)).sum()
                m2 = (sigma * x2 / ((t + sigma) ** 2)).sum()
                res2 = (x2 / ((1 + sigma / t) ** 2)).sum() + res_x2
                alpha = gamma / (m2 + 1e-5)
                beta = (N - gamma) / (res2 + 1e-5)
                t_ = alpha / beta
                # print(abs(t_ - t) / t)
                # evidence = D / 2.0 * np.log(alpha) \
                #            + N / 2.0 * np.log(beta) \
                #            - 0.5 * np.sum(np.log(alpha + beta * sigma)) \
                #            - beta / 2.0 * res2 \
                #            - alpha / 2.0 * m2 \
                #            - N / 2.0 * np.log(2 * np.pi)
                # evidence /= N
                if abs(t_ - t) / t <= 1e-3:  # abs(t_ - t) <= 1e-5 or abs(1 / t_ - 1 / t) <= 1e-5:
                    break

            # if abs(t_ - t) / t > 1e-3:
            #     print('Not converged!')

            evidence = D / 2.0 * np.log(alpha) \
                       + N / 2.0 * np.log(beta) \
                       - 0.5 * np.sum(np.log(alpha + beta * sigma)) \
                       - beta / 2.0 * res2 \
                       - alpha / 2.0 * m2 \
                       - N / 2.0 * np.log(2 * np.pi)
            evidence /= N
            m = 1.0 / (t + sigma) * s * x
            m = (vh.T @ m).reshape(-1)
            evidences.append(evidence)
            self.alphas.append(alpha)
            self.betas.append(beta)
            self.ms.append(m)
        self.ms = np.stack(self.ms)
        return np.mean(evidences)

    _fit = _fit_fixed_point
    # _fit = _fit_icml

    def fit(self, f: np.ndarray, y: np.ndarray, add_intercept=False):
        """
        :param f: [N, F], feature matrix from pre-trained model
        :param y: target labels.
            For classification, y has shape [N] with element in [0, C_t).
            For regression, y has shape [N, C] with C regression-labels

        :return: LogME score (how well f can fit y directly)
        """
        if self.fitted:
            warnings.warn('re-fitting for new data. old parameters cleared.')
            self.reset()
        else:
            self.fitted = True

        if add_intercept:
            f = np.hstack((np.ones((f.shape[0], 1)), f))

        f = f.astype(np.float64)
        if self.regression:
            y = y.astype(np.float64)
            if len(y.shape) == 1:
                y = y.reshape(-1, 1)

        return self._fit(f, y)

    def predict(self, f: np.ndarray):
        """
        :param f: [N, F], feature matrix
        :return: prediction, return shape [N, X]
        """
        if not self.fitted:
            raise RuntimeError("not fitted, please call fit first")
        f = f.astype(np.float64)
        logits = f @ self.ms.T
        if self.regression:
            return logits
        return np.argmax(logits, axis=-1)
